/* ========================================================================= */
/**
 * @file `(file-name-nondirectory (buffer-file-name))`
 *
 * @copyright
 * Copyright (c) `(format-time-string "%Y")` `user-mail-address`
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifndef __WLMAKER_`(upcase (file-name-nondirectory (file-name-sans-extension (buffer-file-name))))`_H__
#define __WLMAKER_`(upcase (file-name-nondirectory (file-name-sans-extension (buffer-file-name))))`_H__

#ifdef __cplusplus
extern "C" {
#endif  // __cplusplus

$0

#ifdef __cplusplus
}  // extern "C"
#endif  // __cplusplus

#endif  // __WLMAKER_`(upcase (file-name-nondirectory (file-name-sans-extension (buffer-file-name))))`_H__
/* == End of `(file-name-nondirectory (buffer-file-name))` ================= */
