/* Doxygen extra stylesheet. Make it look like a wlmaker window. */

body {
    background-color: #505080;
    margin: 0;
}

body, table, div, p, dl {
    font: 12pt Helvetica,sans-serif;
}

div.contents {
    background-color: #a6a6b6;
    border-color: black;
    border: 1px solid;
    color: black;
    filter: drop-shadow(4px 4px 4px black);
    margin-left: auto;
    margin-right: auto;
    margin-top: 0px;
    padding-bottom: 20px;
    padding-left: 20px;
    padding-right: 20px;
    width: 960px;
}

div.fragment {
    background-color: #202020;
    border: 0px none;
    color: silver;
    padding: 8px;
}

div.header {
    background-color: #505080;
    background-image: linear-gradient(10deg, #000010, #202070);
    border-bottom: 1px solid #101038;
    border-left: 1px solid #808088;
    border-right: 1px solid #101038;
    border-top: 1px solid #808088;
    filter: drop-shadow(4px 4px 4px black);
    margin-bottom: 0px;
    margin-left: auto;
    margin-right: auto;
    margin-top: 20px;
    padding: 0px;
    width: 998px;
}

div.headertitle {
    background-image: url('primitive_minimize_icon.png'), url('primitive_close_icon.png');
    background-position: left, right;
    background-repeat: no-repeat, no-repeat;
    border: 0px none;
    margin: 0px;
    padding-bottom: 0px;
    padding-left: 22px;
    padding-right: 22px;
    padding-top: 0px;
}

div.title {
    border-bottom: 0px none;
    border-left: 2px groove #808088;
    border-right: 2px groove #808088;
    border-top: 0px none;
    color: white;
    font-size: 14px;
    height: 22px;
    margin: 0px;
    padding-left: 4px;
}

/* Title elements */
div.wlmtitle { font-size: 2.5em; }
span.wlmprefix { color: #ffbc01; font-weight: bold; }
img.wlmlogo {
    height: 1.25em;
    width: 1.25em;
    vertical-align: -0.2em;
}

h1 { font-size: 1.4em; }
h2 { font-size: 1.2em; }
h3 { font-size: 1.1em; }

div.line {
    font: 12pt Monospace, fixed;
    line-height: normal;
}

/* Colorization */
.contents a:visited { color: #505080; }
a { color: #3c3c60; text-decoration: underline; }
a:hover { text-decoration: underline dashed; }
a:visited { color: #505080; }
code { background-color: #9898a0; }
div.line { font-size: 12pt; min-height: 14pt; }
span.charliteral { color: lightskyblue; }
span.comment { color: gray; }
span.keyword { color: lightgreen; }
span.keywordtype { color: limegreen; }
span.lineno { color: gray; background-color: #202020; border-right: 2px solid gray; }
span.stringliteral { color: lightskyblue; }

/* hides titlearea and the main navigation. */
#top { display: none; }
hr.footer { display: none; }
.footer { display: none; }
